

class Serializer:

	def __init__(self, destination = None, protocol = 2):
		self.destination = destination
		if type(destination) == str:
			self.format = destination.split('.')[-1]
			if self.format == 'pkl':
				import cPickle as serializer
				if protocol:
					self.protocol = protocol
			elif self.format == 'json':
				import json as serializer
		elif protocol:
			import cPickle as serializer
			self.protocol = protocol
		else:
			import json as serializer
		self.serializer = serializer


	def save(self, data):
		if self.destination:
			with open(self.destination, 'wb') as file:
				if hasattr(self,'protocol'):
					self.serializer.dump(data, file, protocol = self.protocol)
				else:
					self.serializer.dump(data, file)
				file.close()
		else:
			if hasattr(self, 'protocol'):
				self.serial = self.serializer.dumps(data, protocol = self.protocol)
			else:
				self.serial = self.serializer.dumps(data)


	def load(self):
		if self.destination:
			with open(self.destination, 'rb') as file:
				self.data = self.serializer.load(file)
				file.close()
		else:
			self.data = self.serializer.loads(self.serial)


if __name__ == '__main__':
	from zaider import moduleProgramScript
	exec( moduleProgramScript )


from time import time


SETTINGS = {
	'trace'   : False,
	'inplace' : True,
	'stamps'  : [],
	'headers' : [],
	'logger'  : [],
	'counter' : 0,
}


# class Meta(type):
# 	@staticmethod
# 	def __call__( header = '', logger = None ):
# 		if SETTINGS['trace']:
# 			SETTINGS['stamps'].append(time())
# 			SETTINGS['headers'].append(header)
# 			if SETTINGS['counter'] != 0:
# 				entry = "{!s:>4}  |  {:.3f}  |  {}".format( SETTINGS['counter'], SETTINGS['stamps'][-1] - SETTINGS['stamps'][-2], SETTINGS['headers'][-2] )
# 				if SETTINGS['inplace']:
# 					print(entry)
# 				if logger != None:
# 					SETTINGS['logger'] = logger
# 				SETTINGS['logger'].append(entry)
# 			SETTINGS['counter'] +=1


class Tracker:
	# __metaclass__ = Meta


	def __new__(cls, header='', logger=None):
		if SETTINGS['trace']:
			SETTINGS['stamps'].append(time())
			SETTINGS['headers'].append(header)
			if SETTINGS['counter'] != 0:
				entry = "{!s:>4}  |  {:.3f}  |  {}".format(SETTINGS['counter'],
				                                           SETTINGS['stamps'][-1] - SETTINGS['stamps'][-2],
				                                           SETTINGS['headers'][-2])
				if SETTINGS['inplace']:
					print(entry)
				if logger != None:
					SETTINGS['logger'] = logger
				SETTINGS['logger'].append(entry)
			SETTINGS['counter'] += 1


	@staticmethod
	def off():
		SETTINGS['trace'] = False

	@staticmethod
	def on():
		SETTINGS['trace'] = True

	@staticmethod
	def isOn():
		return SETTINGS['trace']

	@staticmethod
	def log():
		print('\n'.join( SETTINGS['logger'] ))


_STOPWATCH = {
	'last_duration': 0,
	'start':0,
	'duration': 0,
	'end': 0
}

class Stopwatch:

	@staticmethod
	def last():
		return _STOPWATCH['last_duration']

	@staticmethod
	def start():
		_STOPWATCH['last_duration'] = _STOPWATCH['duration']
		_STOPWATCH['start'] = time()
		_STOPWATCH['duration'] = 0
		_STOPWATCH['end'] = 0

	@staticmethod
	def end():
		_STOPWATCH['end'] = time()
		_STOPWATCH['duration'] = _STOPWATCH['end'] - _STOPWATCH['start']


if __name__ == '__main__':
	from zaider import moduleProgramScript
	exec( moduleProgramScript )


import numpy as np


def cross_ref(x1, x2=None, reftype='holistic', new_axis=-1, ref_axis=-1, ):
	if x2 is None:
		x2 = x1

	if reftype == 'holistic':
		s1 = x1.shape[ref_axis]
		s2 = x2.shape[ref_axis]
		x1 = np.repeat(np.expand_dims(x1, new_axis), s2, axis=new_axis)
		x2 = np.repeat(np.expand_dims(x2, new_axis), s1, axis=new_axis)
		x1 = np.swapaxes(x1, new_axis - 1, ref_axis)
	elif reftype == 'strict upper triangle':
		pass

	return x1, x2

def test(spefarg=None,*args,**kwargs):
	"""
	Test function
	:param spefarg: specific arg for test
	:param args: args for test
	:param kwargs: kwargs for test
	:return:
	"""
	print(spefarg,type(spefarg))

	for arg in args:
		print(arg,type(arg))

	for k,arg in kwargs.items():
		print(k,arg,type(k),type(arg))

if __name__ == '__main__':
	from zaider import moduleProgramScript
	exec(moduleProgramScript)
	# from zaider import Program
	# Program(globals(), locals()).run()


import numpy as np
import sys

def rec(n):
	"""
	Returns the nth Recaman number
	"""
	c=0
	s=set()
	for i in range(n):
		c -= i
		if c in s or c < 0:
			c += 2*i
		s.add(c)
	return c


def rec2(n):
	"""
	Returns a generator with the first n numbers of the Recaman sequence
	"""
	c=0
	s=set()
	for i in range(n):
		c -= i
		if c in s or c < 0:
			c += 2*i
		yield c
		s.add(c)


def fib(n):
	"""
	Returns the nth Fibonacci number
	"""
	a = b = 1
	for _ in range(n-1):
		a, b = b, a+b
	return a


def fib2(n):
	"""
	Returns a generator with the first n numbers of the Fibonacci sequence
	"""
	a = b = 1
	for _ in range(n-1):
		yield a
		a, b = b, a+b
	yield a


def div2(n=10e4, primes_mask = True, composite_mask = False, highly_composite_mask = True ):
	x = np.arange(1,n+1)
	ma = np.fromfunction(lambda *shape: shape[-2] > shape[-1], (n, n))
	xx = np.ma.array( np.lib.stride_tricks.as_strided(x,shape=(n,n),strides=(x.itemsize,0)), mask=ma )
	y = np.sum( ( x % xx ) == 0, axis = 0 ).view(np.ndarray)
	yield x
	yield y
	if primes_mask: yield y == 2
	if composite_mask: yield y > 2
	if highly_composite_mask:
		ma = np.fromfunction(lambda *shape: shape[-2] >= shape[-1], (n, n))
		yy = np.ma.array(np.lib.stride_tricks.as_strided(y, shape=(n, n), strides=(y.itemsize, 0)), mask=ma)
		yield ~np.any(y <= yy, axis=0)


def primes2(n=10e4):
	x,_,m = div2(n,True,False,False)
	return x[m]


def hcn2(n=10e4):
	"""
	Highly Composite Numbers
	:param n:
	:return:
	"""
	x,y,m = div2(n,False,False,True)
	return x[m],y[m]


if __name__=='__main__':
	from zaider import moduleProgramScript
	exec(moduleProgramScript)


from zaider.diag import Tracker
import sys

import importlib as importer
import os
from time import time
from pprint import pprint

ZAIDERPATH = os.environ['Zaider']
sys.path.append( ZAIDERPATH + '/envs/py3/Lib/' )
sys.path.append( ZAIDERPATH + '/envs/py3/Lib/site-packages/' )

def impZd(name,package=''):
	if package == '': package = name.split('.')[0]
	zaiderenv = ZAIDERPATH + '/envs/py3/Lib/'
	try:
		sp = importer.import_module( name, zaiderenv+package+'.py' )
		return sp
	except:
		zaiderenv += 'site-packages/'
		return importer.import_module( name, zaiderenv+package+'.py' )


def findZd(name,package=''):
	if package == '': package = name.split('.')[0]
	zaiderenv = ZAIDERPATH + 'envs/py3/Lib/'
	try:
		sp = importer.util.find_spec( name, zaiderenv+package )
		return sp
	except:
		zaiderenv += 'site-packages/'
		return importer.util.find_spec( name, zaiderenv+package )


class Program:
	def __init__(self, globalV, localV):
		# pprint(globalV)
		# pprint(localV)
		self.namespace = [ globalV, localV ]
		args = sys.argv[1:]

		need_help = False
		c = 0
		for arg in args:
			if arg in ('-h','--help'):
				need_help = True
			elif arg in ('-t','--tracker'):
				Tracker.on()

			elif arg.startswith('-p:'):
				self.protocol = arg[3:]
			elif arg.startswith('--protocol:'):
				self.protocol = arg[11:]

			elif arg.startswith('-to:'):
				self.output_to = arg[4:]

			elif arg.startswith('-o:'):
				self.out_type = eval(arg[3:], globalV, localV)
			elif arg.startswith('--output:'):
				self.out_type = eval(arg[8:], globalV, localV)

			elif arg.startswith('-om:'):
				self.out_method = arg[4:]
			elif arg.startswith('--output_method:'):
				self.out_method = arg[17:]

			else:
				break
			c += 1

		try:
			self.func = eval(args[c], globalV, localV)
			c+=1
		except:
			try:
				self.func = eval('main', globalV, localV)
			except:
				if c > len(args)-1:
					sys.exit(f'Cannot find main function')
				else:
					sys.exit(f'Cannot find either {args[c]} or main function')

		if need_help:
			help(self.func)
			sys.exit()

		self.args = [ eval(arg, globalV, localV) for arg in args[c:] if arg.find('=')==-1 ]
		self.kwargs = { eval(a[0],globalV, localV):eval(a[1],globalV, localV) for a in [ arg.split('=',1) for arg in args[c:] if arg.find('=')>0 ] }

		if Tracker.isOn():
			print(f'\nFunction to be executed: {self.func.__name__}')


		if Tracker.isOn():
			print('\nRunning program...\n')
			start = time()

		# for i in range(len(self.args)):
		# 	arg = self.args[i]
		# 	if type(arg) == str:
		# 		try:
		# 			if '"' not in arg:
		# 				self.args[i] = f'"{arg}"'
		# 			elif "'" not in arg:
		# 				self.args[i] = f"'{arg}'"
		# 			else:
		# 				raise
		# 		except:
		# 			print(f'cannot add quotes to argument: {arg}')

		# call = 'self.command( {} )'.format(', '.join(self.args))
		try:
			# out = eval( call )
			out = self.func(*self.args,**self.kwargs)
			# print('out: ',out)
		except Exception as E:
			raise E
			# raise Exception('{}\n{}'.format(E,call))
		try:
			elapsed = time() - start
			cmd_time = f'\nFunction elapse time: {elapsed}'
		except:
			pass


		if out is not None:
			if hasattr(self, 'out_type'):
				if type(out) != self.out_type:
					try:
						out = self.out_type(out)
					except:
						if Tracker.isOn():
							print(f'\nCould not convert output to {self.out_type}\n')

			print('\nOutput:\n')
			if hasattr(self,'out_method'):
				if self.out_method in ('pp', 'pprint'):
					from pprint import pprint
					pprint(out)
				else:
					print(out)

				# else:
				# 	import _pickle as pickle
				# 	print( pickle.dumps(out, eval(self.protocol)) )

		try:
			print(cmd_time)
		except: pass
		# try:
		# 	elapsed = time() - start
		# 	print('\nProgram elapse time: {}'.format(elapsed))
		# except: pass

moduleProgramScript = """
from zaider import Program
Program( globals(), locals() )
"""


from setuptools import setup

setup(
	name='zaider',
	version='0.0.1',
	description='Zaider Technologies packages',
	packages=['zaider'],
	package_dir={'':'src'},

	author='Costin Leonte',
	author_email='costin.leonte@gmail.com',
	url='https://github.com/costin-leonte/zaider'
)




def union_intersecting_sets(sets):
	'''
	:param list_of_sets: list of sets
	:return: list of unioned intersecting sets
	'''
	if sets[0] is not set:
		sets[0] = set(sets[0])
	i = 1
	while i < len(sets):
		if sets[i] is not set:
			sets[i] = set(sets[i])
		j = 0
		while j < i:
			if len(sets[i] & sets[j]) > 0:
				sets[j] = sets[i] | sets[j]
				del sets[i]
				i -= 1
				break
			j += 1
		i += 1
	return sets


if __name__ == '__main__':
	from zaider import moduleProgramScript
	exec( moduleProgramScript )